/*
Copyright (c) 2001 David Drysdale

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
AUTHOR BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN
AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

Except as contained in this notice, the name of the author shall not be
used in advertising or otherwise to promote the sale, use or other dealings
in this Software without prior written authorization from the author.
*/


import java.awt.*;


/*
 *
 * GridLetter
 *
 */
class GridLetter {
    static final String fontName = "Dialog";
    static final int quantLineWidth = 3;
    static final int pointWidth = 2; // must be <= quantLineWidth

    // The following members depend upon the instantiated size
    Font labelFont;
    FontMetrics fm;
    int  quantSize;
    boolean isFocus;
    int hsize;
    int vsize;
    // The top left position of the letter
    int left;
    int top;
    // The vertical delta to move down to get to the font baseline
    int baselineOffset;
    // The owning GridFont instance
    GridFont owner;
    // The owning GridCanvas instance
    GridCanvas canvas;
    char letter;

    // Each grid point may have up to four quantua emanating from it:
    //    /0
    //  x -1
    //  | \2
    //  3
    static final int UP_RIGHT = 0;
    static final int RIGHT = 1;
    static final int DOWN_RIGHT = 2;
    static final int DOWN = 3;
    boolean quanta[][][];

    // Mappings between internal format of letters and bitstring formats
    static boolean inited = false;
    static int quantaBit[][][]; // which bit number corresponds to this quantum
    static int iiForBit[]; // given bit number x, these arrays yield the [ii][jj][kk]
    static int jjForBit[]; // values in the internal format
    static int kkForBit[];
    
    // Initialize mapping tables.  This assumes standard 2x6 letters
    void initTables() {
        if (inited) return;

        if (owner.debugMode) System.out.println("Initialize quanta mappings");
        quantaBit = new int[3][7][4];
        iiForBit = new int[56];
        jjForBit = new int[56];
        kkForBit = new int[56];
        
        iiForBit[ 0] = 0; jjForBit[ 0] = 0; kkForBit[ 0] = 1;
        iiForBit[ 1] = 1; jjForBit[ 1] = 0; kkForBit[ 1] = 1;
        iiForBit[ 2] = 0; jjForBit[ 2] = 1; kkForBit[ 2] = 1;
        iiForBit[ 3] = 1; jjForBit[ 3] = 1; kkForBit[ 3] = 1;
        iiForBit[ 4] = 0; jjForBit[ 4] = 2; kkForBit[ 4] = 1;
        iiForBit[ 5] = 1; jjForBit[ 5] = 2; kkForBit[ 5] = 1;
        iiForBit[ 6] = 0; jjForBit[ 6] = 3; kkForBit[ 6] = 1;
        iiForBit[ 7] = 1; jjForBit[ 7] = 3; kkForBit[ 7] = 1;
        iiForBit[ 8] = 0; jjForBit[ 8] = 4; kkForBit[ 8] = 1;
        iiForBit[ 9] = 1; jjForBit[ 9] = 4; kkForBit[ 9] = 1;
        iiForBit[10] = 0; jjForBit[10] = 5; kkForBit[10] = 1;
        iiForBit[11] = 1; jjForBit[11] = 5; kkForBit[11] = 1;
        iiForBit[12] = 0; jjForBit[12] = 6; kkForBit[12] = 1;
        iiForBit[13] = 1; jjForBit[13] = 6; kkForBit[13] = 1;

        iiForBit[14] = 0; jjForBit[14] = 0; kkForBit[14] = 3;
        iiForBit[15] = 1; jjForBit[15] = 0; kkForBit[15] = 3;
        iiForBit[16] = 2; jjForBit[16] = 0; kkForBit[16] = 3;
        iiForBit[17] = 0; jjForBit[17] = 1; kkForBit[17] = 3;
        iiForBit[18] = 1; jjForBit[18] = 1; kkForBit[18] = 3;
        iiForBit[19] = 2; jjForBit[19] = 1; kkForBit[19] = 3;
        iiForBit[20] = 0; jjForBit[20] = 2; kkForBit[20] = 3;
        iiForBit[21] = 1; jjForBit[21] = 2; kkForBit[21] = 3;
        iiForBit[22] = 2; jjForBit[22] = 2; kkForBit[22] = 3;
        iiForBit[23] = 0; jjForBit[23] = 3; kkForBit[23] = 3;
        iiForBit[24] = 1; jjForBit[24] = 3; kkForBit[24] = 3;
        iiForBit[25] = 2; jjForBit[25] = 3; kkForBit[25] = 3;
        iiForBit[26] = 0; jjForBit[26] = 4; kkForBit[26] = 3;
        iiForBit[27] = 1; jjForBit[27] = 4; kkForBit[27] = 3;
        iiForBit[28] = 2; jjForBit[28] = 4; kkForBit[28] = 3;
        iiForBit[29] = 0; jjForBit[29] = 5; kkForBit[29] = 3;
        iiForBit[30] = 1; jjForBit[30] = 5; kkForBit[30] = 3;
        iiForBit[31] = 2; jjForBit[31] = 5; kkForBit[31] = 3;

        iiForBit[32] = 0; jjForBit[32] = 1; kkForBit[32] = 0;
        iiForBit[33] = 1; jjForBit[33] = 1; kkForBit[33] = 0;
        iiForBit[34] = 0; jjForBit[34] = 2; kkForBit[34] = 0;
        iiForBit[35] = 1; jjForBit[35] = 2; kkForBit[35] = 0;
        iiForBit[36] = 0; jjForBit[36] = 3; kkForBit[36] = 0;
        iiForBit[37] = 1; jjForBit[37] = 3; kkForBit[37] = 0;
        iiForBit[38] = 0; jjForBit[38] = 4; kkForBit[38] = 0;
        iiForBit[39] = 1; jjForBit[39] = 4; kkForBit[39] = 0;
        iiForBit[40] = 0; jjForBit[40] = 5; kkForBit[40] = 0;
        iiForBit[41] = 1; jjForBit[41] = 5; kkForBit[41] = 0;
        iiForBit[42] = 0; jjForBit[42] = 6; kkForBit[42] = 0;
        iiForBit[43] = 1; jjForBit[43] = 6; kkForBit[43] = 0;

        iiForBit[44] = 0; jjForBit[44] = 0; kkForBit[44] = 2;
        iiForBit[45] = 1; jjForBit[45] = 0; kkForBit[45] = 2;
        iiForBit[46] = 0; jjForBit[46] = 1; kkForBit[46] = 2;
        iiForBit[47] = 1; jjForBit[47] = 1; kkForBit[47] = 2;
        iiForBit[48] = 0; jjForBit[48] = 2; kkForBit[48] = 2;
        iiForBit[49] = 1; jjForBit[49] = 2; kkForBit[49] = 2;
        iiForBit[50] = 0; jjForBit[50] = 3; kkForBit[50] = 2;
        iiForBit[51] = 1; jjForBit[51] = 3; kkForBit[51] = 2;
        iiForBit[52] = 0; jjForBit[52] = 4; kkForBit[52] = 2;
        iiForBit[53] = 1; jjForBit[53] = 4; kkForBit[53] = 2;
        iiForBit[54] = 0; jjForBit[54] = 5; kkForBit[54] = 2;
        iiForBit[55] = 1; jjForBit[55] = 5; kkForBit[55] = 2;

        // Now produce the inverse table
        if (owner.debugMode) System.out.println("now the inverse tables");
        for (int ii=0; ii<3; ii++) {
            for (int jj=0; jj<7; jj++) {
                for (int kk=0; kk<4; kk++) {
                    quantaBit[ii][jj][kk] = -1;
                }
            }
        }
        for (int ll=0; ll<56; ll++) {
            quantaBit[iiForBit[ll]][jjForBit[ll]][kkForBit[ll]] = ll;
        }

        inited = true;
    }

    // Constructor
    public GridLetter(GridFont gf, GridCanvas cv, Graphics g, int x, int y, int hz, int vz, char let) {
        owner = gf; // do this first so tracing works
        initTables();
        if (owner.debugMode) System.out.println("construct letter "+let+" at ("+x+", "+y+")");
        canvas = cv;
        left = x; top = y; 
        hsize = hz; vsize = vz;
        letter = let;
        // Horizontally, we need to fit in
        //   i) the describing letter OR
        //   ii) two quanta, together with their caps
        // So first, find a font that fits horizontally
        int fontsize=hsize+1;
        int pointsize=hsize+1;
        fm = null;
        while (fontsize  > hsize) {
            pointsize--;
            labelFont = new Font(fontName, Font.PLAIN, pointsize);
            fm = g.getFontMetrics(labelFont);
            fontsize = fm.getMaxAdvance();
            if (fontsize == -1) fontsize=hsize+1; // no font metric - try next font down
        }

        // Next, figure out the horizontal quantum size
        // |<---hsize------->|
        //    xxxxxxxxxxxxx
        //   xx     x     xx
        //   xx     x     xx
        //    xxxxxxxxxxxxx
        //  ^              ^
        //  |            room for cap of line width
        //  single space at either side
        int hquantsize = (hsize - quantLineWidth - 1 - 2) / owner.numHDivs; 

        // Vertically, we need to fit in 
        //  i) the describing letter AND
        //  ii) 6 quanta 
        fontsize = fm.getHeight();
        int vquantsize = (vsize - quantLineWidth - 1 - 2 - fontsize) / owner.numVDivs;
            
        // keep reducing the size of the font until we have at least 3 pels for
        // each quantum, and also until the label part takes up less than 20% of 
        // the vertical space available 
        while ((vquantsize < 3) || ((double)fontsize > (0.2 * (double)vsize))) {
            pointsize--;
            labelFont = new Font(fontName, Font.PLAIN, pointsize);
            fm = g.getFontMetrics(labelFont);
            fontsize = fm.getHeight();
            vquantsize = (vsize - quantLineWidth - 1 - 2 - fontsize) / owner.numVDivs;
        }
        quantSize = Math.min(hquantsize, vquantsize);

        // Adjust hsize and vsize to reflect the overall quantSize
        hsize = quantSize*owner.numHDivs+1+2+2*(quantLineWidth/2);
        vsize = quantSize*owner.numVDivs+1+2+2*(quantLineWidth/2)+fontsize;
        baselineOffset = fm.getAscent();
        isFocus = false;

        // Initialize the collection of quanta
        quanta = new boolean[owner.numHDivs+1][owner.numVDivs+1][4];
        clear();
    }
    public void mouseDown(int x, int y) {
        int ii = (x - left - 2 - (quantLineWidth/2))/quantSize;
        int jj = (y - top  - 2 - (quantLineWidth/2))/quantSize; 
        int deltax = (x - left - 2 - (quantLineWidth/2))%quantSize;
        int deltay = (y - top  - 2 - (quantLineWidth/2))%quantSize; 
        if ((ii < -1) || (jj < -1) || (ii > owner.numHDivs) || (jj > owner.numVDivs)) return;
        // We want to cope with clicks just a little to the left/above the
        // leftmost/topmost square
        if (ii == -1) {
            ii = 0; deltax -= quantSize;
        }
        if (jj == -1) {
            jj = 0; deltay -= quantSize;
        }

        if (owner.debugMode) {
            String out = "x="+x+" y="+y+" left="+left+" top="+top+" quantSize="+quantSize+
                " ii="+ii+" jj="+jj+" deltax="+deltax+" deltay="+deltay;
            System.out.println(out);
        }

        int thresh = 1 + quantLineWidth/2;

        if ((Math.abs(deltax) <= thresh) && (jj<owner.numVDivs)) {
            quanta[ii][jj][DOWN] = !quanta[ii][jj][DOWN];
        }
        if ((Math.abs(quantSize - deltax) <= thresh)  && 
            (ii < owner.numHDivs) && (jj<owner.numVDivs)){
            quanta[ii+1][jj][DOWN] = !quanta[ii+1][jj][DOWN];
        }
        if ((Math.abs(deltay) <= thresh) && (ii<owner.numHDivs)){
            quanta[ii][jj][RIGHT] = !quanta[ii][jj][RIGHT];
        }
        if ((Math.abs(quantSize - deltay) <= thresh)  && 
            (ii<owner.numHDivs) && (jj < owner.numVDivs)){
            quanta[ii][jj+1][RIGHT] = !quanta[ii][jj+1][RIGHT];
        }
        // If the click was strictly too far left/up, then it cannot trigger a 
        // diagonal toggle
        if ((deltax >= 0) && (deltay >= 0)) {
            if ((Math.abs(deltax - deltay) <= thresh) && 
                (ii < owner.numHDivs) && (jj < owner.numVDivs)) {
                quanta[ii][jj][DOWN_RIGHT] = !quanta[ii][jj][DOWN_RIGHT];
            }
            if ((Math.abs(quantSize - (deltax + deltay)) <= thresh) && 
                (ii < owner.numHDivs) && (jj < owner.numVDivs)) {
                quanta[ii][jj+1][UP_RIGHT] = !quanta[ii][jj+1][UP_RIGHT];
            }
        }
    }
    public void getQuanta(boolean copy[][][]) {
        for (int ii=0; ii<=owner.numHDivs; ii++) {
            for (int jj=0; jj<=owner.numVDivs; jj++) {
                for (int kk=0; kk<4; kk++) {
                    copy[ii][jj][kk] = quanta[ii][jj][kk];
                }
            }
        }
    }
    public void setQuanta(boolean copy[][][]) {
        for (int ii=0; ii<=owner.numHDivs; ii++) {
            for (int jj=0; jj<=owner.numVDivs; jj++) {
                for (int kk=0; kk<4; kk++) {
                    quanta[ii][jj][kk] = copy[ii][jj][kk];
                }
            }
        }
    }

    // The following two methods only work for standard sized grid letters,
    // that is 2x6
    public long getEncoding() {
        long code = 0;
        for (int ii=0; ii<=owner.numHDivs; ii++) {
            for (int jj=0; jj<=owner.numVDivs; jj++) {
                for (int kk=0; kk<4; kk++) {
                    if (quanta[ii][jj][kk]) {
                        if (quantaBit[ii][jj][kk] != -1) {
                            code |= (1L << (55-quantaBit[ii][jj][kk]));
                        } else {
                            // should never get here!
                        }
                    }
                }
            }
        }
        return code;
    }
    public void setEncoding(long code) {
        for (int ll=0; ll<56; ll++) {
            quanta[iiForBit[ll]][jjForBit[ll]][kkForBit[ll]] = 
                ((code & (1L << (55-ll))) != 0L) ? true : false;
        }
        draw(canvas.getGraphics());
    }
    public String getEncodingString() {
        long code = getEncoding();
        String trunc = Long.toHexString(code);
        // need to left pad with 0s till we reach 14 digits
        while (trunc.length() < 14) {
            trunc = "0" + trunc;
        }
        return(new String(letter + " " + trunc.toUpperCase()));
    }
    public void setEncoding(String codeString) {
        long code = 0;
        try {
            code = Long.parseLong(codeString, 16);
        } catch (NumberFormatException e) {}
        setEncoding(code);
    }
        
    public int getHSize() {return hsize;}
    public int getVSize() {return vsize;}
    public void gainedFocus() {
        boolean oldFocus = isFocus;
        isFocus = true;
        if (oldFocus != isFocus) draw(canvas.getGraphics());
    }
    public void lostFocus() {
        boolean oldFocus = isFocus;
        isFocus = false;
        if (oldFocus != isFocus) draw(canvas.getGraphics());
    }
    public void clear() {
        for (int ii=0; ii<=owner.numHDivs; ii++) {
            for (int jj=0; jj<=owner.numVDivs; jj++) {
                for (int kk=0; kk<4; kk++) {
                    quanta[ii][jj][kk] = false;
                }
            }
        }
    }
    boolean[][][] makeCopy() {
       boolean[][][]old = new boolean[owner.numHDivs+1][owner.numVDivs+1][4];
        for (int ii=0; ii<=owner.numHDivs; ii++) {
            for (int jj=0; jj<=owner.numVDivs; jj++) {
                for (int kk=0; kk<4; kk++) {
                    old[ii][jj][kk] = quanta[ii][jj][kk];
                }
            }
        }
        return old;
    }
    public void hflip() {
        boolean old[][][] = makeCopy();
        int N = owner.numVDivs;
        for (int jj=0; jj<=N; jj++) {
            if (jj > 0) quanta[0][jj][0] = old[1][jj-1][2]; else quanta[0][jj][0] = false;
                        quanta[0][jj][1] = old[1][jj]  [1];
            if (jj < N) quanta[0][jj][2] = old[1][jj+1][0]; else quanta[0][jj][2] = false;
                        quanta[0][jj][3] = old[2][jj]  [3];
            
            if (jj > 0) quanta[1][jj][0] = old[0][jj-1][2]; else quanta[1][jj][0] = false;
                        quanta[1][jj][1] = old[0][jj]  [1];
            if (jj < N) quanta[1][jj][2] = old[0][jj+1][0]; else quanta[1][jj][2] = false;
                        quanta[1][jj][3] = old[1][jj]  [3];
            
                        quanta[2][jj][0] = false;
                        quanta[2][jj][1] = false;
                        quanta[2][jj][2] = false;
                        quanta[2][jj][3] = old[0][jj]  [3];
        }

    }
    public void vflip() {
        boolean old[][][] = makeCopy();
        int N = owner.numHDivs;
        for (int ii=0; ii<=N; ii++) {
            quanta[ii][0][0] = old[ii][6][2]; // false
            quanta[ii][0][1] = old[ii][6][1];
            quanta[ii][0][2] = old[ii][6][0];
            quanta[ii][0][3] = old[ii][5][3]; 

            quanta[ii][1][0] = old[ii][5][2];
            quanta[ii][1][1] = old[ii][5][1];
            quanta[ii][1][2] = old[ii][5][0];
            quanta[ii][1][3] = old[ii][4][3];

            quanta[ii][2][0] = old[ii][4][2];
            quanta[ii][2][1] = old[ii][4][1];
            quanta[ii][2][2] = old[ii][4][0];
            quanta[ii][2][3] = old[ii][3][3];

            quanta[ii][3][0] = old[ii][3][2];
            quanta[ii][3][1] = old[ii][3][1];
            quanta[ii][3][2] = old[ii][3][0];
            quanta[ii][3][3] = old[ii][2][3];

            quanta[ii][4][0] = old[ii][2][2];
            quanta[ii][4][1] = old[ii][2][1];
            quanta[ii][4][2] = old[ii][2][0];
            quanta[ii][4][3] = old[ii][1][3];

            quanta[ii][5][0] = old[ii][1][2];
            quanta[ii][5][1] = old[ii][1][1];
            quanta[ii][5][2] = old[ii][1][0];
            quanta[ii][5][3] = old[ii][0][3];

            quanta[ii][6][0] = old[ii][0][2];
            quanta[ii][6][1] = old[ii][0][1];
            quanta[ii][6][2] = old[ii][0][0];
            quanta[ii][6][3] = false;
        }
    }
    public void rotate() {
        vflip();
        hflip();
    }
    public void draw(Graphics g) {
        if (owner.debugMode) System.out.println("draw letter  "+letter);
        // Clear out the drawing area first
        g.clearRect(left, top, hsize, vsize);

        // Draw this letter in blue if it is the current focus; black otherwise
        if (isFocus) {
            g.setColor(Color.blue);
        } else {
            g.setColor(Color.black);
        }

        // First draw the grid of points, if necessary
        if (owner.showGrid()) {
            for (int ii=0; ii<=owner.numHDivs; ii++) {
                for (int jj=0; jj<=owner.numVDivs; jj++) {
                    g.fillOval(left + 1 + (quantLineWidth/2) + ii*quantSize - pointWidth,
                               top  + 1 + (quantLineWidth/2) + jj*quantSize - pointWidth,
                               2 * pointWidth,
                               2 * pointWidth);
                }
            }
        }

        // Next, draw the quanta which are set
        int x1,y1,x2,y2;
        for (int ii=0; ii<=owner.numHDivs; ii++) {
            for (int jj=0; jj<=owner.numVDivs; jj++) {
                x1=left+1+(quantLineWidth/2) + ii*quantSize;
                y1=top +1+(quantLineWidth/2) + jj*quantSize;
                for (int kk=0; kk<4; kk++) {
                    if (quanta[ii][jj][kk]) {
                        switch (kk) {
                        case UP_RIGHT:
                            x2 = x1 + quantSize;
                            y2 = y1 - quantSize;
                            break;
                        case RIGHT:
                            x2 = x1 + quantSize;
                            y2 = y1;
                            break;
                        case DOWN_RIGHT:
                            x2 = x1 + quantSize;
                            y2 = y1 + quantSize;
                            break;
                        case DOWN:
                        default:
                            x2 = x1;
                            y2 = y1 + quantSize;
                            break;
                        }
                        g.drawLine(x1, y1, x2, y2);                        
                    }
                }
            }
        }

        // Finally, draw the labelling letter
        g.setFont(labelFont);
        char ca[] = new char[1]; ca[0] = letter;
        g.drawString(new String(ca), 
                      left + Math.max(0,((hsize - fm.charWidth(letter))/2)), 
                      top + owner.numVDivs*quantSize + quantLineWidth + 2 + baselineOffset);
    }
}
